#ifndef DGX_SHAPES_3D_H
#define DGX_SHAPES_3D_H

#include <engine/psx.h>
#include <engine/3d/entity.h>

/* WARNING: CIRCLE_LINEF2 must be a multiple of 3. */
#define CIRCLE_LINEF2		15

/*
 * NOTE: structures prepended with 'Ass' indicate that they
 * are used in a particular asset format. This usually means
 * not supporting individual rotations but instead exclusively
 * supporting the rotation of its30 (the asset) amongst
 * other things. This is to save on space. The normal versions
 * of these structures are to be used sparingly (e.g. debugging).
 */

/*** asset shapes ***/
struct AssQuad
{
	LINE_F2			prim[4];
	VECTOR			translation;

	struct
	{
		enum
		{
			CXC_SLOPE_FLAT	= 0x00,
			CXC_SLOPE_X	= (1<<0),
			CXC_SLOPE_Z	= (1<<1)
		}direction;

		int		is_flipped;

		SVECTOR		v[4];
	}global_coords;

	/* FIXME: DEBUG */
	TILE_8			aabb_corner[2];

	short			aabb_x[2];
	short			aabb_y[2];
	short			aabb_z[2];
};

struct AssCuboid
{
	struct AssQuad		quad[6];

	/*
	 * FIXME: might just be better to use the aabb's from quad[0]
	 * instead of being redundant
	 */
	short			aabb_x[2];
	short			aabb_y[2];
	short			aabb_z[2];
};

/*** normal shapes ***/
struct Quad
{
	POLY_F4			prim;
	SVECTOR			rotation;	/* rotation relative to the camera */
	SVECTOR			rotation_model;	/* rotation in model space */
	VECTOR			translation;
	SVECTOR			global_pos;
	unsigned short		is_coplanar;

	/* if !is_coplanar */
	struct
	{
		SVECTOR		v[4];
	}global_coords;

	/* if is_coplanar */
	unsigned short		size;		/* size of _half_ a plane's edge */
};

struct Cuboid
{
	VECTOR			translation;
	SVECTOR			offset;		/* rename this to global_pos already */
	SVECTOR			rotation;
	int			size_x, size_y, size_z;

	/* debug members */
	LINE_F4			left;
	LINE_F2			left_;

	LINE_F4			right;
	LINE_F2			right_;

	LINE_F4			top;
	LINE_F2			top_;

	LINE_F4			bottom;
	LINE_F2			bottom_;
};

struct Circle3D
{
	struct Entity3D		ent; /* scale is radius */
	LINE_F2			lines[CIRCLE_LINEF2];
	SVECTOR			vertex[CIRCLE_LINEF2];
	VECTOR			unit[2]; /* b-hat and c-hat */

	enum
	{
		SOLID,
		DISCONNECTED,
		DASHED
	}visual;

	enum
	{
		CIRCLE_HORIZONTAL,
		CIRCLE_VERTICAL,
		CIRCLE_CUSTOM
	}orientation;
};

/* right cylinder is assumed */
struct Cylinder
{
	int			r, h;
	SVECTOR			p[2];
};

#define CYLINDERGFX_BODY_NLINES		5
#define CYLINDERGFX_BODY_NSEGMENTS	3

struct CylinderGFX
{
	struct Cylinder		cyl;
	struct Entity3D		ent;
	VECTOR			unit; /* a-hat */

	struct Circle3D		endc[2];
	TILE			endp[2];
	LINE_F2			body[CYLINDERGFX_BODY_NLINES][CYLINDERGFX_BODY_NSEGMENTS];
};

struct Capsule
{
	struct Cylinder		cyl;
	int			r[2];
};

#endif

